# author: Nicolas Tessore <n.tessore@ucl.ac.uk>
# license: MIT
'''main module of the cosmology package'''

__version__ = '2022.10.9'

__all__ = [
    'Cosmology',
]

from .background import Cosmology
